from typing import Tuple
import numpy as np
from numpy import pi as π
from sadie.agents.exceptions import NoTargetError

EPSILON: float = 1e-6


class TargetingMixin:
    """
    Targeting management mixin. Provides for target tracking, retrieval and azimuth/distance calculations to target.
    """

    def __init_subclass__(cls, **kwargs):
        cls._target = (None, None)

    @property
    def target(self) -> Tuple[float, float]:
        """
        Returns the current target of the targetable agent.

        :return: the tuple of current target coordinates
        """
        return self._target

    @target.setter
    def target(self, value):
        raise ValueError("Cannot set target directly. Please use the set_ methods to target your object.")

    def clear_target(self):
        """
        Resets the targetable object to untargeted.
        """
        self._target = (None, None)

    def set_absolute_target(self, x: float, y: float):
        """
        Sets the target of the agent to `(x, y)`.

        :param x: absolute x position to set target to
        :param y: absolute y position to set target to
        """
        self._target = (x, y)

    def set_relative_target(self, x: float, y: float):
        """
        Sets the target of the agent relative to its current position by `(x, y)`.

        :param x: x-offset of the target
        :param y: y-offset of the target
        """
        self._target = (self.x + x, self.y + y)

    def set_polar_target(self, theta: float, r: float):
        """
        Sets the target of the agent relative to its current position by reference to an azimuth and a distance.

        :param theta: azimuth to target point
        :param r: distance to target point
        """
        self._target = (self.x + np.cos(theta) * r,
                        self.y + np.sin(theta) * r)

    @property
    def target_distance(self) -> float:
        """
        Returns distance of the agent from its current target.

        :return: distance of the agent from its current target
        :raise NoTargetError: the agent is not targeted
        """
        if self._target == (None, None):
            raise NoTargetError
        else:
            return self.distance_from((self._target[0], self._target[1]))

    @property
    def target_azimuth(self) -> float:
        """
        Returns the azimuth of the agent towards its current target.

        :return: azimuth of the agent towards its current target
        :raise NoTargetError: the agent is not targeted
        """
        if self._target == (None, None):
            raise NoTargetError
        else:
            theta = np.arctan2(self._target[1] - self.y, self._target[0] - self.x)
            if theta >= 0:
                return theta
            else:
                return 2 * π + theta

    @property
    def is_on_target(self) -> bool:
        """
        Returns `True` if the object is within a tolerance margin (set as `EPSILON`, by default :math:`10^{-6}`) from
        its destination point.

        :return: `True` if the object is within the tolerance margin of the target, else `False`
        :raise NoTargetError: the agent is not targeted
        """
        if self._target == (None, None):
            raise NoTargetError
        else:
            return self.target_distance <= EPSILON



class NoTargetError(Exception):
    def __init__(self, *args, **kwargs):
        self.message = "The current agent does not have a target, therefore no target azimuth or distance can be " \
                       "calculated."



from typing import Optional, Callable
from scipy import stats

from sadie.agents.exceptions import NoTargetError
from sadie.agents.spatial import TargetableAgent, AgentStates
import numpy as np
from numpy import pi as π


class BaseWalker(TargetableAgent):
    """
    The `BaseWalker` is a base class that implements the basic notion of a _walker_: an agent that sets and follows
    a pattern of sequentially assigned targets autonomously.

    The `BaseWalker` does not implement waiting (i.e. it's always on the move) and it does not exhibit extraneously
    sensitive behaviour (i.e. it does not react to the space it is moving across), and has a constant velocity.
    """
    def retarget(self):
        self.set_polar_target(np.random.uniform(0, 2*π), np.random.randint(1, 100))

    def update(self):
        if self.target == (None, None):
            self.retarget()
        elif self.is_on_target:
            self.retarget()
        else:
            self.move()


class WaitingUniformWalker(BaseWalker):
    def __init__(self, x_init: float, y_init: float, velocity: float = 1.0, wait_transition_probability: float = 0.4):
        super(WaitingUniformWalker, self).__init__(x_init=x_init, y_init=y_init, velocity=velocity)
        self.wait_transition_probability = wait_transition_probability

    def update(self):
        if self.target == (None, None):
            self.retarget()
        elif self.is_on_target:
            if np.random.random() > self.wait_transition_probability:
                self.retarget()
            else:
                self.wait()
        else:
            self.move()


class UniformLevyRandomWalker(BaseWalker):
    """
    The Uniform-Lévy random walker determines its trip distances according to the Lévy distribution

    .. math::

        f(x) =  \\frac{e^{- \\frac{1}{2x}}}{\\sqrt{2\\pi x^3}}

    and the azimuth of each target is drawn from a uniform distribution of :math:`(0, 2\\pi]`.
    """

    def retarget(self):
        self.set_polar_target(np.random.uniform(0, 2*π), stats.levy().rvs())


class BoundedUniformLevyRandomWalker(BaseWalker):
    def __init__(self,
                 x_init: float,
                 y_init: float,
                 velocity: float = 1.0,
                 wait_transition_probability: Optional[float] = None,
                 bounding_distribution: stats.rv_continuous = stats.norm,
                 **kwargs):
        super(BoundedUniformLevyRandomWalker, self).__init__(x_init=x_init, y_init=y_init, velocity=velocity)
        self.wait_transition_probability = wait_transition_probability
        self.bounding_distribution = bounding_distribution
        self.kwargs = kwargs

    def update(self):
        if self.target == (None, None):
            self.retarget()
        elif self.is_on_target:
            self.retarget()
        elif self.trip_distance >= self.bounding_distribution.rvs(**self.kwargs):
            self.retarget()
            self.reset_trip_odometer()
        else:
            self.move()


class HomesickLevyWalker(BaseWalker):
    """
    Implements the Homesick Lévy walk described by `Fujihara and Miwa (2014) <https://arxiv.org/abs/1408.0427>`_. This
    walk has a home location, to which the walker returns, and a homesick probability :math:`\\alpha`, which describes
    the likelihood that the agent will set its home point as the next target.

    Note that this follows the Fujihara and Miwa paper's conceptualisation of homesickness, i.e. a turn to the home
    location is only possible at the end of a trip. Therefore, given :math:`\\alpha`, a walker will make `\\alpha^{-1}`
    trips before setting course for its home location. For a Lévy walker that can abandon a trip in progress, try
    `RapidHomesickLevyWalker`.
    """
    def __init__(self, x_init: float, y_init: float, velocity: float = 1.0, alpha: float = 0.2):
        super(HomesickLevyWalker, self).__init__(x_init=x_init, y_init=y_init, velocity=velocity)
        self.home_x, self.home_y = x_init, y_init
        self.alpha = alpha

    def update(self):
        if self.target == (None, None):
            self.retarget()
        elif self.is_on_target:
            if np.random.random() > self.alpha:
                self.retarget()
            else:
                self.set_absolute_target(self.home_x, self.home_y)
        else:
            self.move()


class RapidHomesickLevyWalker(BaseWalker):
    """
    Implements a variant of the Homesick Lévy walk described by `Fujihara and Miwa (2014)
    <https://arxiv.org/abs/1408.0427>`_ implemented in `HomesickLevyWalker`. Like the original `HomesickLevyWalker`,
    the `RapidHomesickLevyWalker` is executing a homesick Lévy walk around its point of origin. Unlike the original
    `HomesickLevyWalker`, however, the `RapidHomesickLevyWalker` determines whether to retarget for home not every time
    it has reached a location but at every movement step. Consequently, the meaning of `\\alpha` is different:
    a walker will travel a distance of `\\alpha^{-1}` before setting course for its home location.
    """
    def __init__(self, x_init: float, y_init: float, velocity: float = 1.0, alpha: float = 0.2):
        super(RapidHomesickLevyWalker, self).__init__(x_init=x_init, y_init=y_init, velocity=velocity)
        self.home_x, self.home_y = x_init, y_init
        self.alpha = alpha

    def update(self):
        if self.target == (None, None):
            self.retarget()
        elif self.is_on_target:
            if np.random.random() > self.alpha:
                self.retarget()
            else:
                self.set_absolute_target(self.home_x, self.home_y)
        else:
            if self.target == (self.home_x, self.home_y):
                self.move()
            else:
                if np.random.random() > self.alpha:
                    self.move()
                else:
                    self.set_absolute_target(self.home_x, self.home_y)



class ModelRunNotCompletedError(Exception):
    def __init__(self, *args, **kwargs):
        self.message = "This model has not finished running. Please wait for the model to finish running before using " \
                       "an export method."



import unittest
import numpy as np
from sadie.models import simple, ModelStatus, AbstractAgent
from sadie.agents.spatial import SpatialAgent

class AbstractModelFunctionalityTest(unittest.TestCase):
    def setUp(self) -> None:
        self.model = simple.SimpleModel()
        self.agt = SpatialAgent(0, 0)

    def test_add_agent(self):
        self.assertEqual(self.model.agents, [])
        self.model.add_agent(self.agt)
        self.assertEqual(len(self.model.agents), 1)

    def test_agents_listing(self):
        self.model.add_agent(self.agt)
        self.assertEqual(self.model.agents[0], self.agt)
        self.assertIsInstance(self.model.agents, list)

    def test_agents_setting(self):
        agt1, agt2 = SpatialAgent(1, 1), SpatialAgent(2, 2)
        self.model.agents = [agt1, agt2]
        self.assertEqual(self.model.agents[0], agt1)
        self.assertEqual(self.model.agents[1], agt2)

    def test_zero_starting_time(self):
        self.assertEqual(self.model.time, 0)

    def test_time_protection(self):
        with self.assertRaises(ValueError):
            self.model.time = 1

    def test_state_before_run(self):
        self.assertEqual(self.model.state, ModelStatus.NOT_RUN)

    def test_state_protection(self):
        with self.assertRaises(ValueError):
            self.model.state = ModelStatus.DONE

    def test_population_with_arbitrary_class(self):
        class TestObject(SpatialAgent):
            pass

        rx, ry = np.random.randint(-100, 100, 2)

        m = simple.SimpleModel()
        m.populate(10, TestObject, x_init = rx, y_init = ry)

        for i in m.agents:
            self.assertIsInstance(i, TestObject)

        self.assertEqual(len(m.agents), 10)
        self.assertEqual(m.agents[0].x, rx)
        self.assertEqual(m.agents[0].y, ry)




from uuid import uuid4
from abc import ABC, abstractmethod


class AbstractAgent(ABC):
    """
    The `AbstractAgent` class is the abstract base class from which agents in general inherit. The `AbstractAgent`
    incorporates the three main properties of an agent:

    * an identifier (typically a UUID4, but can be manually provided as a keyword argument called `id`),
    * the `update()` method, which embodies a time step, and
    * the `report()` method, which allows the agent to report its current state.

    Any class inheriting from `AbstractAgent` absolutely *must* implement the `update()` and `report()` methods.
    """

    def __init__(self, *args, **kwargs):
        if "id" in kwargs:
            self.id = kwargs["id"]
        else:
            self.id = str(uuid4())

    @abstractmethod
    def update(self):
        """
        Updates the agent.
        """
        pass

    @abstractmethod
    def report(self):
        """
        Draws a report from the agent.
        """
        pass



import unittest
from sadie.agents.spatial import MovingSpatialAgent
from random import randint, random
import numpy as np
from numpy import pi as π


class TestMovingSpatialAgent(unittest.TestCase):
    def test_odometer(self):
        agent = MovingSpatialAgent(randint(-100, 100), randint(-100, 100))
        self.assertEqual(agent.distance_traveled, 0)
        agent.move_by(1, 0)
        self.assertEqual(agent.distance_traveled, 1)
        agent.move_by(1, 0)
        self.assertEqual(agent.distance_traveled, 2)
        xoffs, yoffs = randint(-100, 100) * random(), randint(-100, 100) * random()
        agent.move_by(xoffs, yoffs)
        self.assertAlmostEqual(agent.distance_traveled, 2 + np.sqrt(xoffs ** 2 + yoffs ** 2))

    def test_move_to(self):
        xinit, yinit, xdelta, ydelta = randint(-100, 100), \
                                       randint(-100, 100), \
                                       randint(-100, 100) * random(), \
                                       randint(-100, 100) * random()
        agent = MovingSpatialAgent(x_init=xinit, y_init=yinit)
        agent.move_to(xdelta, ydelta)
        self.assertEqual(agent.position, (xdelta, ydelta))

    def test_move_by(self):
        xinit, yinit, xdelta, ydelta = randint(-100, 100), \
                                       randint(-100, 100), \
                                       randint(-100, 100) * random(), \
                                       randint(-100, 100) * random()
        agent = MovingSpatialAgent(x_init=xinit, y_init=yinit)
        agent.move_by(xdelta, ydelta)
        self.assertAlmostEqual(agent.position, (xinit + xdelta, yinit + ydelta))
        self.assertAlmostEqual(agent.distance_traveled, np.sqrt(xdelta ** 2 + ydelta ** 2))

    def test_move_p(self):
        xinit, yinit = randint(-100, 100), randint(-100, 100)
        theta, r = np.random.uniform(0, 2*π), randint(-100, 100)
        agent = MovingSpatialAgent(x_init=xinit, y_init=yinit)
        agent.move_p(theta, r)
        self.assertAlmostEqual(agent.position, (xinit + np.cos(theta) * r, yinit + np.sin(theta) * r))
        self.assertAlmostEqual(agent.distance_traveled, r)

    def test_successive_move_accumulation_in_odometer(self):
        xinit, yinit = randint(-100, 100), randint(-100, 100)
        thetas, rs = np.random.uniform(0, 2*π, 100), np.random.uniform(1, 100, 100)
        agent = MovingSpatialAgent(x_init=xinit, y_init=yinit)
        for i, j in zip(thetas, rs):
            agent.move_p(i, j)
        self.assertAlmostEqual(agent.distance_traveled, rs.sum())



import unittest
try:
    import mock
except ImportError:
    from unittest import mock

import numpy as np
from numpy import pi as π
from sadie.agents.spatial import AgentStates
from sadie.agents.walkers import BaseWalker


class TestBaseWalker(unittest.TestCase):
    def test_retargeting_behaviour_on_creation(self):
        xinit, yinit = np.random.uniform(-100, 100, 2)

        w = BaseWalker(x_init=xinit, y_init=yinit)

        self.assertEqual(w.x, xinit, msg="Starting position must be the initialising position.")
        self.assertEqual(w.y, yinit, msg="Starting position must be the initialising position.")
        self.assertEqual(w.target, (None, None), msg="At initialisation, the walker must be untargeted.")
        self.assertEqual(w.state, AgentStates.HALTED, msg="At initialisation, the walker must be in the HALTED "
                                                            "state.")

        w.update()

        self.assertEqual(w.x, xinit, msg="The first update must not change the x coordinates before the move.")
        self.assertEqual(w.y, yinit, msg="The first update must not change the y coordinates before the move.")
        self.assertIsNotNone(w.target[0], msg="The target x must be set after the first update.")
        self.assertIsNotNone(w.target[1], msg="The target y must be set after the first update.")
        wx1, wy1 = w.x, w.y
        tx1, ty1 = w.target

        w.update()

        self.assertEqual(w.state, AgentStates.MOVING)
        self.assertNotEqual(w.position, (xinit, yinit), msg="The second update should move the agent's position.")
        self.assertNotEqual(w.distance_traveled, 0, msg="At the second update, the agent must have a positive "
                                                          "nonzero distance traveled.")
        self.assertEqual(w.target, (tx1, ty1), msg="At the second update, the target should remain the same.")
        self.assertNotEqual((w.x, w.y), (wx1, wy1), msg="At the second update, the agent must have moved from its "
                                                          "position in the previous state.")

    @mock.patch("numpy.random.uniform", return_value=π)
    @mock.patch("numpy.random.randint", return_value=4)
    def test_random_walk_behaviour(self, m_azimuth, m_distance):
        w = BaseWalker(0, 0)
        self.assertEqual(w.x, 0)
        self.assertEqual(w.y, 0)
        self.assertEqual(w.target, (None, None))
        w.update()
        self.assertNotEqual(w.target, (None, None))
        w.update()
        self.assertAlmostEqual(w.target[0], np.cos(π) * 4)
        self.assertAlmostEqual(w.target[1], np.sin(π) * 4)

        for i in range(3):
            w.update()

        self.assertTrue(w.is_on_target)
        self.assertAlmostEqual(w.distance_traveled, 4)
